/* =============================================================================
 *       Filename:  Extractor.cpp
 *    Description:  Extractor class
 *        Created:  11/15/11 00:43:19
 *
 *         Author:  Derek Tingle, derek.tingle@gmail.com
 * ===========================================================================*/

#include <stdio.h>
#include <string>
#include <math.h>
#include <iostream>
#include <xtract/libxtract.h>

#include "Extractor.h"

using namespace std;

Extractor::Extractor(int c_buff_size){
  buff_size=c_buff_size*NUM_CHANS*ZPF;
  buff=NULL;
  buff=new SAMPLE[buff_size];
  clear_buff();


  // prep spectrum params
  // since samplerate can change the first param is modified in extract_mfccs
  m_spect_params[1]=(float)XTRACT_MAGNITUDE_SPECTRUM;
  m_spect_params[2]=0; // not getting DC

  // the xtract_mel_filter struct is declared in libxtract.h
  m_mel_filter = new xtract_mel_filter;
  // number of mfccs
  m_mel_filter->n_filters=20;
  m_mel_filter->filters = new SAMPLE *[m_mel_filter->n_filters];

  // stores spectrum
  m_spect=new float[buff_size];
  // stores array used by XTRACT_MFCC
  m_tmp_mfccs=new SAMPLE[buff_size/2];
  // stores buffer for use by get_mfcc
  m_mfcc_buff=new SAMPLE[buff_size];
  // stores final mfcc
  mfccs=new double[m_mel_filter->n_filters];


  // allocate arrays for m_mel_filter struct
  for (int i=0; i<m_mel_filter->n_filters;++i){
    m_mel_filter->filters[i]=new SAMPLE[buff_size/2];
  }
}

Extractor::~Extractor(){
  delete [] buff;
  delete [] m_spect;
  delete [] m_tmp_mfccs;
  delete [] mfccs;
  // delete arrays inside m_mel_filter struct
  for (int i=0; i<m_mel_filter->n_filters;++i){
    delete [] m_mel_filter->filters[i];
  }
  delete m_mel_filter;
}


/* ===  FUNCTION  ==============================================================
 *         Name:  get_mfccs
 *  Description:  finds mfccs
 * ===========================================================================*/
void Extractor::get_mfccs(int sr){
  // clean up arrays
  memset(m_spect,0,buff_size*sizeof(float));
  memset(m_tmp_mfccs,0,buff_size/2*sizeof(SAMPLE));
  memset(mfccs,0,m_mel_filter->n_filters*sizeof(double));
  memset(m_mfcc_buff,0,buff_size*sizeof(SAMPLE));

  for (int i=0;i<buff_size;++i){
    m_mfcc_buff[i]=buff[i];
  }

  monofy();
  apply_window();


  // since this can change...
  m_spect_params[0]=(float)sr/buff_size;


  // get spectrum
  xtract[XTRACT_SPECTRUM](m_mfcc_buff,buff_size,m_spect_params,m_spect);

  // initialize mfcc params stored in m_mel_filter struct
  xtract_init_mfcc(buff_size/2,sr/2.0f,XTRACT_EQUAL_GAIN,20.0f,20000.0f,
      m_mel_filter->n_filters,m_mel_filter->filters);

  // get the mfccs
  xtract[XTRACT_MFCC](m_spect,buff_size/2,m_mel_filter,m_tmp_mfccs);

  for (int i=0; i<m_mel_filter->n_filters;++i){
    mfccs[i]=(double)m_tmp_mfccs[i];
  }

  return;
}

/* ===  FUNCTION  ==============================================================
 *         Name:  apply_window
 *  Description:  applies a hanning window to the global buffer
 *                (shamelessly copied from Ge Wang's chuck_fft code)
 * ===========================================================================*/
void Extractor::apply_window(){
  int length=buff_size/ZPF/NUM_CHANS;
  double pi, phase=0, delta;

  delta = 2 * PI / (double) length;

  for(int i = 0; i < length; i++ ){
    m_mfcc_buff[i] = m_mfcc_buff[i]*(float)(0.5 * (1.0 - cos(phase)));
    phase += delta;
  }

  return;
}

/* ===  FUNCTION  ==============================================================
 *         Name:  monofy
 *  Description:  make a buffer with multiple interleaved channels into a mono
 *                buffer by adding the channels for each frame.
 * ===========================================================================*/
void Extractor::monofy(){
  int mono_size=buff_size/NUM_CHANS/ZPF;

  // sum up each channel and divide by number of channels
  for (int i=0;i<buff_size;++i){
    // if index is within mono length, sum channels
    if (i<mono_size){
      m_mfcc_buff[i]=m_mfcc_buff[i*NUM_CHANS]/NUM_CHANS;
      for (int j=1; j<NUM_CHANS;++j){
        m_mfcc_buff[i]+=m_mfcc_buff[(i*NUM_CHANS)+j]/NUM_CHANS;
      }
    }
    // if not, these vals have all been summed, so zero 'em out
    else{
      m_mfcc_buff[i]=0;
    }
  }
  
  return;
}

/* ===  FUNCTION  ==============================================================
 *         Name:  detect_silence
 *  Description:  looks for silent frames
 * ===========================================================================*/
int Extractor::detect_silence(){
  for (int i=0;i<buff_size/ZPF;++i){
    if (fabs(buff[i])>0)
      return 0;
  }
  return 1;
}


/* ===  FUNCTION  ==============================================================
 *         Name:  clear_buff
 *  Description:  memsets buff
 * ===========================================================================*/
void Extractor::clear_buff(){
  memset(buff,0,buff_size*sizeof(SAMPLE));
  return;
}
