/* =============================================================================
 *       Filename:  extract_feats.cpp
 *    Description:  extracts features from audio files and stores them in db
 *        Created:  11/15/11 00:43:19
 *
 *         Author:  Derek Tingle, derek.tingle@gmail.com
 * ===========================================================================*/

#include <stdio.h>
#include <string>
#include <exception>
#include <math.h>
#include <time.h>

#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/progress.hpp>

#include <soci/soci.h>
#include <soci/soci-backend.h>
#include <soci/sqlite3/soci-sqlite3.h>
#include <sndfile.h>
#include <xtract/libxtract.h>

#include "../Definitions.h"
#include "Extractor.h"

using namespace std;
namespace fs = boost::filesystem;

// function prototypes
void analyze_file(fs::path f);
bool get_track_id(string track_name, int * id);
void extract_mfccs(int track_id,int offset,int sr);
void extract_zcr(SAMPLE * outtie,int track_id,int offset);
void monofy();
void init_globals();
void destroy_globals();
int detect_silence();

//SAMPLE * zero_pad(SAMPLE * buff, int *buff_size, int zpf);

// setup sql
soci::session g_sql;
//  soci::session g_sql(soci::sqlite3, "/Users/dt/code/ccrma/256a/project/data/test.db");


/* ===  FUNCTION  ==============================================================
 *         Name:  analyze_file
 *  Description:  takes a filename, extracts features, and adds them to the db
 * ===========================================================================*/
void analyze_file(fs::path f){
  if (DEBUG)
    cout << endl << "analyzing " << f.c_str() << endl;

  string name=f.stem().string() + f.extension().string();
  int sr, track_id=-1;
  bool exists=get_track_id(name,&track_id);

  if (exists){
    if (DEBUG)
      cout << "\tskipping it" << endl;
    return;
  }

  // setup libsndfile
  SF_INFO *song_info=new SF_INFO;
  song_info -> format = 0;

  if (DEBUG)
    cout << "\topening" << endl;

  SNDFILE* song_file = sf_open(f.c_str(), SFM_READ,song_info);
  if (song_info->channels!=NUM_CHANS){
    cerr << "Error: " << f.string() << " has " << song_info->channels << " channels" << endl;
    exit(1);
  }
  sr = song_info->samplerate;

//// debug using a 2 channel sine wave at nyquist rate
//  for (int i=0; i<BUFF_SIZE*NUM_CHANS;++i){
//    x.buff[i]=((i%4==0) || (i%4==1)) ? 1 : -1;
//  }

  // TODO: go through and replace all values of BUFF_SIZE with buff_size
  //       make sure it looks legit
  //       add some debug output
  
  // added this loop to populate a table with features from many buffer sizes
  int buff_size;
  // buff_size takes on values: 512,1024,2048,4096,8192
  for (int i=0;i<5;++i){
    buff_size=2048*pow(2,i);
    soci::transaction tr(g_sql);
    Extractor x(buff_size);

    sf_seek(song_file,0,SEEK_SET);

    // number of frames
    int num_buffs = (int) ceil(song_info->frames/(float)buff_size);
    // hop (since we're windowing)
    int hop_size = buff_size/2;
    int num_hops= num_buffs*2-1;

    // i found that using a transaction and individual inserts is just as fast
    // as bulk inserts and a little easier to code 

    if (DEBUG)
        cout << "\treading through file" << endl;

    // loop through each frame
    for (int i=0; i<num_hops; ++i){
      if (DEBUG && (i%100==0))
    //      cout << ".";

      x.clear_buff();
      sf_readf_float(song_file, x.buff, (sf_count_t) buff_size);

      // backup by hop_size for next iteration
      sf_seek(song_file,-1*hop_size,SEEK_CUR);

      // if the frame is silent, skip it
      if(x.detect_silence()){
        cerr << "*" ;
        continue;
      }

      // exttract mfccs and insert them into db
      x.get_mfccs(sr);

      // check to see if there's anything close to what we're about to insert
      int close=0;
      try{
        g_sql << "select  count(*) from r_mfccs join r_mfccs_data on r_mfccs.id=r_mfccs_data.id and buff_size=:buff_size and min_mfcc1>=:min_m1 and max_mfcc1<=:max_m1 and min_mfcc2>=:min_m2 and max_mfcc2<=:max_m2 and min_mfcc3>=:min_m3 and max_mfcc3<=:max_m3 and min_mfcc4>=:min_m4 and max_mfcc4<=:max_m4 and min_mfcc5>=:min_m5 and max_mfcc5<=:max_m5;",
              soci::use(buff_size),
              soci::use(x.mfccs[0]-2),soci::use(x.mfccs[0]+2),
              soci::use(x.mfccs[1]-2),soci::use(x.mfccs[1]+2),
              soci::use(x.mfccs[2]-2),soci::use(x.mfccs[2]+2),
              soci::use(x.mfccs[3]-2),soci::use(x.mfccs[3]+2),
              soci::use(x.mfccs[4]-2),soci::use(x.mfccs[4]+2),
              soci::into(close);
      }
      catch( exception const & e){
        cerr << e.what() << endl;
        sf_close(song_file);
        exit(1);
      }

     // if (close>0){
     //   cerr << ".";
     //   continue;
     // }

      // prepare blob for write to db

      // insert into db
      try{
        g_sql << "insert into r_mfccs_data (track_id,frame_start,buff_size) values(:id,:offset,:buff_size);", soci::use(track_id), soci::use(i*hop_size),soci::use(buff_size);

        g_sql << "insert into r_mfccs (id, min_mfcc1,max_mfcc1,min_mfcc2,max_mfcc2,min_mfcc3,max_mfcc3,min_mfcc4,max_mfcc4,min_mfcc5,max_mfcc5) values((select last_insert_rowid() from r_mfccs_data),:min_m1,:max_m1,:min_m2,:max_m2,:min_m3,:max_m3,:min_m4,:max_m4,:min_m5,:max_m5);",
              soci::use(x.mfccs[0]-.5),soci::use(x.mfccs[0]+.5),
              soci::use(x.mfccs[1]-.5),soci::use(x.mfccs[1]+.5),
              soci::use(x.mfccs[2]-.5),soci::use(x.mfccs[2]+.5),
              soci::use(x.mfccs[3]-.5),soci::use(x.mfccs[3]+.5),
              soci::use(x.mfccs[4]-.5),soci::use(x.mfccs[4]+.5);
      }
      catch( exception const & e){
        cerr << e.what() << endl;
        sf_close(song_file);
        exit(1);
      }
    }

    tr.commit();
    if (DEBUG)
      cout << "\n\tinserting feats for buff_size="<< buff_size << endl;

  }
  delete song_info;
  sf_close(song_file);

  return;
}

/* ===  FUNCTION  ==============================================================
 *         Name:  get_track_id
 *  Description:  looks up a track in the db and sets the value of the passed
 *                in int pointer to the id. if the track isn't found, create it.
 *                returns a boolean indicating if the entry already existed 
 * ===========================================================================*/
bool get_track_id(string track_name,int * id){
  bool exists;
  g_sql << "select id from tracks where name=:name",soci::use(track_name),soci::into(*id);
  if (*id>0){
    exists=true;
    if (DEBUG)
      cout << "\tfound track in the db with id=" << *id << endl;
  }
  else{
    g_sql << "insert into tracks (name) values(:name)",soci::use(track_name);
    g_sql << "select id from tracks where name=:name",soci::use(track_name),soci::into(*id);
    exists=false;
    if (DEBUG)
      cout << "\tinserted track into the db. id=" << *id << endl;
  }
  return exists; 
}

/* ===  FUNCTION  ==============================================================
 *         Name:  main
 *  Description:  loops through each wav file in passed-in directory. analyzes
 *                track and inserts values into db.
 * ===========================================================================*/
int main(int argc, char* argv[]){
  try{
    g_sql.open(soci::sqlite3,DB_PATH); 


  fs::path wav_dir;
  if (argc == 2){
    wav_dir = fs::system_complete(fs::path(argv[1]));
  }
  else{
    cerr << "\nUsage: extract_feats [directory]" << endl;
    exit(1);
  }
  if (!fs::is_directory(wav_dir)){
    cerr << "Error: " << wav_dir << " doesn't appear to be a directory" << endl;
    exit(1);
  }
    
  fs::directory_iterator end_itr;
  for (fs::directory_iterator lser(wav_dir); lser != end_itr; ++lser){
    if (fs::is_regular_file(lser->status())){
      if (lser->path().extension()==".wav"){
        // if the file's a wav, analyze it
        analyze_file(lser->path());
      }
    }
  }
  }
  catch( exception const & e){
    cerr << e.what() << endl;
    exit(1);
  }

  return 0;
}
